#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <dirent.h>
#include <stdlib.h>
#include "rods.h"
#include "rodsLog.h"
#include "rsApiHandler.h"
#include "apiHeaderAll.h"
#include "GlobusDRSVars.h"

void waitTillTaskEnds(pid_t pid) {

	char procfile[64];
	sprintf(procfile, "/proc/%i/exe", pid);
	FILE *file;

	while (1) {

		usleep(100);
		file = fopen(procfile, "r");

		if (NULL == file) break;

		fclose(file);
	}
}

char* readFile(char* fileName) {

	char* buffer = NULL;
	FILE* file = fopen(fileName, "r");

	if (NULL == file) {

		rodsLog(LOG_ERROR, "Error opening file %s", fileName);
		return NULL;
	}
	else {

		int pos = -1;
		char c;

		do {

			c = fgetc (file);
			pos++;

		} while (EOF != c);

		rewind(file);
		buffer = (char*) malloc(pos * sizeof(char));
		pos = 0;
		c = fgetc(file);

		while (EOF != c) {

			buffer[pos++] = c;
			c = fgetc(file);
		}

		fclose(file);
	}

	return buffer;
}

char* getPhysicalPath(char* rodsFile, rsComm_t* rsComm) {

	char* physicalPath = NULL;
	dataObjInfo_t* dataObjInfo = NULL;
	dataObjInp_t dataObjInp;
	memset(&dataObjInp, 0, sizeof(dataObjInp));

	strncpy(dataObjInp.objPath, rodsFile, strlen(rodsFile));

	if (0 == getDataObjInfo(rsComm, &dataObjInp, &dataObjInfo, NULL, 1)) {

		physicalPath = (char*) malloc(strlen(dataObjInfo->filePath) * sizeof(char));
		strncpy(physicalPath, dataObjInfo->filePath, strlen(dataObjInfo->filePath));
	}
	else {

		rodsLog(LOG_ERROR, "getDataObjInfo failed for %s", rodsFile);
	}
	freeAllDataObjInfo(dataObjInfo);

	return physicalPath;
}

char* getPhysicalPathOfNonexistentFile(char* rodsFile, rsComm_t* rsComm) {

	rescGrpInfo_t* rescGrpInfo = (rescGrpInfo_t*) malloc(sizeof(rescGrpInfo));
	memset(&rescGrpInfo, 0, sizeof(rescGrpInfo));
	getRescInfo(rsComm, rsComm->myEnv.rodsDefResource, NULL, &rescGrpInfo);
	char* physFile = (char*) malloc(1088 * sizeof(char));
	memset(physFile, 0, 1088 * sizeof(char));
	strncpy(physFile, rescGrpInfo->rescInfo->rescVaultPath, strlen(rescGrpInfo->rescInfo->rescVaultPath));
	strncat(physFile, rodsFile + strlen(rescGrpInfo->rescInfo->zoneName) + 1, strlen(rodsFile) - strlen(rescGrpInfo->rescInfo->zoneName) - 1);
	freeAllRescGrpInfo(rescGrpInfo);
	return physFile;
}

#ifndef GRIDFTP_IRODS_MODULE
/*
int setUserProxy(rsComm_t* rsComm) {

	if (NULL == getenv("X509_USER_PROXY")) {

		char* physProxy;
		char* rodsProxy = (char*) malloc((strlen(rsComm->myEnv.rodsHome) + strlen(PROXY_LOCATION)) * sizeof(char));
		strncpy(rodsProxy, rsComm->myEnv.rodsHome, strlen(rsComm->myEnv.rodsHome));
		strncat(rodsProxy, PROXY_LOCATION, strlen(PROXY_LOCATION));

		if (NULL != (physProxy = getPhysicalPath(rodsProxy, rsComm))) {

			setenv("X509_USER_PROXY", physProxy, 1);
			free(physProxy);
			free(rodsProxy);
		}
		else {

			rodsLog(LOG_ERROR, "setting user proxy failed");
		}
	}

	return 0;
}
*/
#endif

char* replicationPut(char* rlsUrl, char* options, char* rodsFile, char* LFN, char* destUrl, rsComm_t* rsComm) {

	pid_t pid;
	char* tmpFilename = tmpnam(NULL);
	char* physFile = getPhysicalPath(rodsFile, rsComm);

	if (NULL == options) {

		options = "";
	}

	if (0 > (pid = fork())) {

		rodsLog(LOG_ERROR, "globus-replication-client: fork() failed");
	}
	else if (0 == pid) {

		char exeBuffer[1024];
		char argBuffer[10240];
		sprintf(exeBuffer, "%s/bin/globus-replication-client", GLOBUS_LOCATION);
		sprintf(argBuffer, "-r %s %s put %s %s %s", rlsUrl, options, physFile, LFN, destUrl);
		execl(SHELL_SCRIPT, "", exeBuffer, argBuffer, tmpFilename, NULL);
		rodsLog(LOG_ERROR, "globus-replication-client: execl() failed");
		exit(0);
	}

	waitTillTaskEnds(pid);
	char* result = readFile(tmpFilename);
	unlink(tmpFilename);
	free(physFile);
	return result;
}

char* replicationGet(char* rlsUrl, char* options, char* LFN, char* rodsFile, rsComm_t* rsComm, ruleExecInfo_t* rei) {

	pid_t pid;
	char* tmpFilename = tmpnam(NULL);
	char* physFile = getPhysicalPathOfNonexistentFile(rodsFile, rsComm);

	if (NULL == options) {

		options = "";
	}

	if (0 > (pid = fork())) {

		rodsLog(LOG_ERROR, "globus-replication-client: fork() failed");
	}
	else if (0 == pid) {

		char exeBuffer[1024];
		char argBuffer[10240];
		sprintf(exeBuffer, "%s/bin/globus-replication-client", GLOBUS_LOCATION);
		sprintf(argBuffer, "-r %s %s get %s %s ", rlsUrl, options, LFN, physFile);
		execl(SHELL_SCRIPT, "", exeBuffer, argBuffer, tmpFilename, NULL);
		rodsLog(LOG_ERROR, "globus-replication-client: execl() failed");
		exit(0);
	}

	waitTillTaskEnds(pid);
	char* result = readFile(tmpFilename);
	unlink(tmpFilename);

	dataObjInp_t* dataObjInp;
	dataObjInp = (dataObjInp_t*) malloc(sizeof(dataObjInp_t));
	memset(dataObjInp, 0, sizeof(dataObjInp_t));
	strncpy(dataObjInp->objPath, rodsFile, strlen(rodsFile));

/*	struct stat buf;
	stat(physFile, &buf);
	int size = buf.st_size;
	dataObjInp->dataSize = (rodsLong_t) size;
*/
	addKeyVal(&dataObjInp->condInput, DEF_RESC_NAME_KW, rsComm->myEnv.rodsDefResource);
	addKeyVal(&dataObjInp->condInput, FILE_PATH_KW, physFile);
	rsPhyPathReg (rsComm, dataObjInp);

	free(physFile);
	free(dataObjInp);

	return result;
}

char* replicationCopy(char* rlsUrl, char* options, char* LFN, char* destUrl) {

	pid_t pid;
	char* tmpFilename = tmpnam(NULL);

	if (NULL == options) {

		options = "";
	}

	if (0 > (pid = fork())) {

		rodsLog(LOG_ERROR, "globus-replication-client: fork() failed");
	}
	else if (0 == pid) {

		char exeBuffer[1024];
		char argBuffer[10240];
		sprintf(exeBuffer, "%s/bin/globus-replication-client", GLOBUS_LOCATION);
		sprintf(argBuffer, "-r %s %s copy %s %s", rlsUrl, options, LFN, destUrl);
		execl(SHELL_SCRIPT, "", exeBuffer, argBuffer, tmpFilename, NULL);
		rodsLog(LOG_ERROR, "globus-replication-client: execl() failed");
		exit(0);
	}

	waitTillTaskEnds(pid);
	char* result = readFile(tmpFilename);
	unlink(tmpFilename);
	return result;
}

char* replicationRegister(char* rlsUrl, char* options, char* LFN, char* destUrl) {

	pid_t pid;
	char* tmpFilename = tmpnam(NULL);

	if (NULL == options) {

		options = "";
	}

	if (0 > (pid = fork())) {

		rodsLog(LOG_ERROR, "globus-replication-client: fork() failed");
	}
	else if (0 == pid) {

		char exeBuffer[1024];
		char argBuffer[10240];
		sprintf(exeBuffer, "%s/bin/globus-replication-client", GLOBUS_LOCATION);
		sprintf(argBuffer, "-r %s %s register %s %s", rlsUrl, options, LFN, destUrl);
		execl(SHELL_SCRIPT, "", exeBuffer, argBuffer, tmpFilename, NULL);
		rodsLog(LOG_ERROR, "globus-replication-client: execl() failed");
		exit(0);
	}

	waitTillTaskEnds(pid);
	char* result = readFile(tmpFilename);
	unlink(tmpFilename);
	return result;
}

char* replicationReplicate(char* rlsUrl, char* options, char* LFN, char* destUrl) {

	pid_t pid;
	char* tmpFilename = tmpnam(NULL);

	if (NULL == options) {

		options = "";
	}

	if (0 > (pid = fork())) {

		rodsLog(LOG_ERROR, "globus-replication-client: fork() failed");
	}
	else if (0 == pid) {

		char exeBuffer[1024];
		char argBuffer[10240];
		sprintf(exeBuffer, "%s/bin/globus-replication-client", GLOBUS_LOCATION);
		sprintf(argBuffer, "-r %s %s replicate %s %s", rlsUrl, options, LFN, destUrl);
		execl(SHELL_SCRIPT, "", exeBuffer, argBuffer, tmpFilename, NULL);
		rodsLog(LOG_ERROR, "globus-replication-client: execl() failed");
		exit(0);
	}

	waitTillTaskEnds(pid);
	char* result = readFile(tmpFilename);
	unlink(tmpFilename);
	return result;
}

char* replicationDelete(char* rlsUrl, char* options, char* LFN, char* destUrl) {

	pid_t pid;
	char* tmpFilename = tmpnam(NULL);

	if (NULL == options) {

		options = "";
	}

	if (0 > (pid = fork())) {

		rodsLog(LOG_ERROR, "globus-replication-client: fork() failed");
	}
	else if (0 == pid) {

		char exeBuffer[1024];
		char argBuffer[10240];
		sprintf(exeBuffer, "%s/bin/globus-replication-client", GLOBUS_LOCATION);
		sprintf(argBuffer, "-r %s %s delete %s %s", rlsUrl, options, LFN, destUrl);
		execl(SHELL_SCRIPT, "", exeBuffer, argBuffer, tmpFilename, NULL);
		rodsLog(LOG_ERROR, "globus-replication-client: execl() failed");
		exit(0);
	}

	waitTillTaskEnds(pid);
	char* result = readFile(tmpFilename);
	unlink(tmpFilename);
	return result;
}

int msiGlobusReplicationPut(
		msParam_t* P_rlsUrl,
		msParam_t* P_options,
		msParam_t* P_rodsFile,
		msParam_t* P_LFN,
		msParam_t* P_destUrl,
		msParam_t* P_out,
		ruleExecInfo_t* rei) {

	char* rlsUrl = (char*) P_rlsUrl->inOutStruct;
	char* options = (char*) P_options->inOutStruct;
	char* rodsFile = (char*) P_rodsFile->inOutStruct;
	char* LFN = (char*) P_LFN->inOutStruct;
	char* destUrl = (char*) P_destUrl->inOutStruct;
	char* out;

	setUserProxy(rei->rsComm);

	if (NULL != (out = replicationPut(rlsUrl, options, rodsFile, LFN, destUrl, rei->rsComm))) {

		fillStrInMsParam(P_out, out);
		free(out);
		rei->status = 1;
	}
	else {

		rodsLog(LOG_ERROR, " msiGlobusReplicationPut(%s, %s, %s, %s, %s) failed", rlsUrl, options, rodsFile, LFN, destUrl);

		out = (char*) malloc(1024 * sizeof(char));
		sprintf(out, " FAILURE: msiGlobusReplicationPut(%s, %s, %s, %s, %s)", rlsUrl, options, rodsFile, LFN, destUrl);
		fillStrInMsParam(P_out, out);
		free(out);

		rei->status = EXEC_CMD_ERROR;
	}

	return rei->status;
}

int msiGlobusReplicationGet(
		msParam_t* P_rlsUrl,
		msParam_t* P_options,
		msParam_t* P_LFN,
		msParam_t* P_rodsFile,
		msParam_t* P_out,
		ruleExecInfo_t* rei) {

	char* rlsUrl = (char*) P_rlsUrl->inOutStruct;
	char* options = (char*) P_options->inOutStruct;
	char* LFN = (char*) P_LFN->inOutStruct;
	char* rodsFile = (char*) P_rodsFile->inOutStruct;
	char* out;

	setUserProxy(rei->rsComm);

	if (NULL != (out = replicationGet(rlsUrl, options, LFN, rodsFile, rei->rsComm, rei))) {

		fillStrInMsParam(P_out, out);
		free(out);
		rei->status = 1;
	}
	else {

		rodsLog(LOG_ERROR, " msiGlobusReplicationGet(%s, %s, %s, %s) failed", rlsUrl, options, LFN, rodsFile);

		out = (char*) malloc(1024 * sizeof(char));
		sprintf(out, " FAILURE: msiGlobusReplicationGet(%s, %s, %s, %s)", rlsUrl, options, LFN, rodsFile);
		fillStrInMsParam(P_out, out);
		free(out);

		rei->status = EXEC_CMD_ERROR;
	}

	return rei->status;
}

int msiGlobusReplicationCopy(
		msParam_t* P_rlsUrl,
		msParam_t* P_options,
		msParam_t* P_LFN,
		msParam_t* P_destUrl,
		msParam_t* P_out,
		ruleExecInfo_t* rei) {

	char* rlsUrl = (char*) P_rlsUrl->inOutStruct;
	char* options = (char*) P_options->inOutStruct;
	char* LFN = (char*) P_LFN->inOutStruct;
	char* destUrl = (char*) P_destUrl->inOutStruct;
	char* out;

	setUserProxy(rei->rsComm);

	if (NULL != (out = replicationCopy(rlsUrl, options, LFN, destUrl))) {

		fillStrInMsParam(P_out, out);
		free(out);
		rei->status = 1;
	}
	else {

		rodsLog(LOG_ERROR, " msiGlobusReplicationCopy(%s, %s, %s, %s) failed", rlsUrl, options, LFN, destUrl);

		out = (char*) malloc(1024 * sizeof(char));
		sprintf(out, " FAILURE: msiGlobusReplicationPut(%s, %s, %s, %s)", rlsUrl, options, LFN, destUrl);
		fillStrInMsParam(P_out, out);
		free(out);

		rei->status = EXEC_CMD_ERROR;
	}

	return rei->status;
}

int msiGlobusReplicationRegister(
		msParam_t* P_rlsUrl,
		msParam_t* P_options,
		msParam_t* P_LFN,
		msParam_t* P_destUrl,
		msParam_t* P_out,
		ruleExecInfo_t* rei) {

	char* rlsUrl = (char*) P_rlsUrl->inOutStruct;
	char* options = (char*) P_options->inOutStruct;
	char* LFN = (char*) P_LFN->inOutStruct;
	char* destUrl = (char*) P_destUrl->inOutStruct;
	char* out;

	setUserProxy(rei->rsComm);

	if (NULL != (out = replicationRegister(rlsUrl, options, LFN, destUrl))) {

		fillStrInMsParam(P_out, out);
		free(out);
		rei->status = 1;
	}
	else {

		rodsLog(LOG_ERROR, " msiGlobusReplicationRegister(%s, %s, %s, %s) failed", rlsUrl, options, LFN, destUrl);

		out = (char*) malloc(1024 * sizeof(char));
		sprintf(out, " FAILURE: msiGlobusReplicationRegister(%s, %s, %s, %s)", rlsUrl, options, LFN, destUrl);
		fillStrInMsParam(P_out, out);
		free(out);

		rei->status = EXEC_CMD_ERROR;
	}

	return rei->status;
}

int msiGlobusReplicationReplicate(
		msParam_t* P_rlsUrl,
		msParam_t* P_options,
		msParam_t* P_LFN,
		msParam_t* P_destUrl,
		msParam_t* P_out,
		ruleExecInfo_t* rei) {

	char* rlsUrl = (char*) P_rlsUrl->inOutStruct;
	char* options = (char*) P_options->inOutStruct;
	char* LFN = (char*) P_LFN->inOutStruct;
	char* destUrl = (char*) P_destUrl->inOutStruct;
	char* out;

	setUserProxy(rei->rsComm);

	if (NULL != (out = replicationReplicate(rlsUrl, options, LFN, destUrl))) {

		fillStrInMsParam(P_out, out);
		free(out);
		rei->status = 1;
	}
	else {

		rodsLog(LOG_ERROR, " msiGlobusReplicationReplicate(%s, %s, %s, %s) failed", rlsUrl, options, LFN, destUrl);

		out = (char*) malloc(1024 * sizeof(char));
		sprintf(out, " FAILURE: msiGlobusReplicationReplicate(%s, %s, %s, %s)", rlsUrl, options, LFN, destUrl);
		fillStrInMsParam(P_out, out);
		free(out);

		rei->status = EXEC_CMD_ERROR;
	}

	return rei->status;
}

int msiGlobusReplicationDelete(
		msParam_t* P_rlsUrl,
		msParam_t* P_options,
		msParam_t* P_LFN,
		msParam_t* P_destUrl,
		msParam_t* P_out,
		ruleExecInfo_t* rei) {

	char* rlsUrl = (char*) P_rlsUrl->inOutStruct;
	char* options = (char*) P_options->inOutStruct;
	char* LFN = (char*) P_LFN->inOutStruct;
	char* destUrl = (char*) P_destUrl->inOutStruct;
	char* out;

	setUserProxy(rei->rsComm);

	if (NULL != (out = replicationDelete(rlsUrl, options, LFN, destUrl))) {

		fillStrInMsParam(P_out, out);
		free(out);
		rei->status = 1;
	}
	else {

		rodsLog(LOG_ERROR, " msiGlobusReplicationDelete(%s, %s, %s, %s) failed", rlsUrl, options, LFN, destUrl);

		out = (char*) malloc(1024 * sizeof(char));
		sprintf(out, " FAILURE: msiGlobusReplicationDelete(%s, %s, %s, %s)", rlsUrl, options, LFN, destUrl);
		fillStrInMsParam(P_out, out);
		free(out);

		rei->status = EXEC_CMD_ERROR;
	}

	return rei->status;
}
